#include <tcl.h>

#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include <string.h>

#include <ps5000aApi.h>

short pico_handle = 0;

// 512 MB (= 2^29 bytes) / 4 channels / 2 byte per channel / 25 (downsample ratio)
#define BUFFER_SIZE ((1<<29)/4/2/25)
int16_t *pico_buffer_A = 0, *pico_buffer_B = 0,
    *pico_buffer_C = 0, *pico_buffer_D = 0;

// see the README why 46875...
#define AWG_SIZE 46875
int16_t *awg_buffer = 0;
double *awg_buffer_double = 0;

char err_msg[100];
PICO_STATUS pico_error;

#define CHECK_PICO_ERROR(message)                       \
    if (pico_error != PICO_OK) {                        \
        snprintf (err_msg, sizeof(err_msg),             \
                  "picoScope: %s - got error: 0x%x",  \
                  message, pico_error);                 \
        Tcl_SetResult (interp, err_msg, TCL_VOLATILE);  \
        return TCL_ERROR;                               \
    }                                                   \

/*
 * Declarations for application-specific command procedures
 */

int picoScopeInitCmd (ClientData clientData, Tcl_Interp *interp,
                      int objc, Tcl_Obj *CONST objv[]);
int picoScopeEndCmd (ClientData clientData, Tcl_Interp *interp,
                     int objc, Tcl_Obj *CONST objv[]);
int picoScopeTriggerCmd (ClientData clientData, Tcl_Interp *interp,
                         int objc, Tcl_Obj *CONST objv[]);
int picoScopeAwgCmd (ClientData clientData, Tcl_Interp *interp,
                     int objc, Tcl_Obj *CONST objv[]);
int picoScopeMeasureCmd (ClientData clientData, Tcl_Interp *interp,
                         int objc, Tcl_Obj *CONST objv[]);

/*
 * Init function is called when the package is loaded.
 */

int Picoscope_Init (Tcl_Interp *interp) {
    /*
     * Initialize the stub table interface, which is
     * described in Chapter 46.
     */

    if (Tcl_InitStubs (interp, "8.1", 0) == NULL) {
        Tcl_SetResult (interp, "problem during Tcl_InitStubs", TCL_VOLATILE);
        return TCL_ERROR;
    }

    /*
     * Register all provided functions
     */

    Tcl_CreateObjCommand (interp, "picoScopeInit", picoScopeInitCmd,
                          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand (interp, "picoScopeEnd", picoScopeEndCmd,
                          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand (interp, "picoScopeTrigger", picoScopeTriggerCmd,
                          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand (interp, "picoScopeAwg", picoScopeAwgCmd,
                          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand (interp, "picoScopeMeasure", picoScopeMeasureCmd,
                          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

    /*
     * Declare that we implement the picoScope package
     * so scripts that do "package require picoScope"
     * can load the library automatically.
     */
    Tcl_PkgProvide (interp, "picoScope", "0.1");
    return TCL_OK;
}


int picoScopeInitCmd (ClientData clientData, Tcl_Interp *interp,
                      int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1) {
        Tcl_WrongNumArgs(interp, 1, objv, "");
        return TCL_ERROR;
    }

    if (pico_handle != 0) {
        // already initialized - abort!
        Tcl_SetResult (interp, "picoScope: already initialized", TCL_VOLATILE);
        return TCL_ERROR;
    }

    // Open first found picoScope, Resolution 14 bit
    // --> maximum resolution with active 4 Channels...
    pico_error = ps5000aOpenUnit (&pico_handle, 0,  PS5000A_DR_14BIT);
    CHECK_PICO_ERROR ("problem during opening the USB connection");

    // Set power supply to external
    if (ps5000aCurrentPowerSource (pico_handle) != PICO_POWER_SUPPLY_CONNECTED) {
        // Wenn noch nicht geschehen - stelle die Stromversorgung auf externe 5V Versorgung um...
        pico_error = ps5000aChangePowerSource (pico_handle, PICO_POWER_SUPPLY_CONNECTED);
        CHECK_PICO_ERROR ("couldn't change to external power supply");
    }

    // Enable all four Channels: DC-coupling, 20V range (except current channel: 5V --> 5A), 0.0V offset
    pico_error = ps5000aSetChannel (pico_handle, PS5000A_CHANNEL_A, 1, PS5000A_DC, PS5000A_5V, 0.0);
    CHECK_PICO_ERROR ("couldn't activate Channel A (current measurement)");

    pico_error = ps5000aSetChannel (pico_handle, PS5000A_CHANNEL_B, 1, PS5000A_DC, PS5000A_20V, 0.0);
    CHECK_PICO_ERROR ("couldn't activate Channel B (voltage L1)");

    pico_error = ps5000aSetChannel (pico_handle, PS5000A_CHANNEL_C, 1, PS5000A_DC, PS5000A_20V, 0.0);
    CHECK_PICO_ERROR ("couldn't activate Channel C (voltage L2)");

    pico_error = ps5000aSetChannel (pico_handle, PS5000A_CHANNEL_D, 1, PS5000A_DC, PS5000A_20V, 0.0);
    CHECK_PICO_ERROR ("couldn't activate Channel D (voltage L3)");

    // activate internal 20 MHz filter -- reduce high frequency noise...
    pico_error = ps5000aSetBandwidthFilter (pico_handle, PS5000A_CHANNEL_A, PS5000A_BW_20MHZ);
    CHECK_PICO_ERROR ("couldn't set 20 MHz filter on Channel A");

    pico_error = ps5000aSetBandwidthFilter (pico_handle, PS5000A_CHANNEL_B, PS5000A_BW_20MHZ);
    CHECK_PICO_ERROR ("couldn't set 20 MHz filter on Channel B");

    pico_error = ps5000aSetBandwidthFilter (pico_handle, PS5000A_CHANNEL_C, PS5000A_BW_20MHZ);
    CHECK_PICO_ERROR ("couldn't set 20 MHz filter on Channel C");

    pico_error = ps5000aSetBandwidthFilter (pico_handle, PS5000A_CHANNEL_D, PS5000A_BW_20MHZ);
    CHECK_PICO_ERROR ("couldn't set 20 MHz filter on Channel D");

    // Trigger - external, rising edge, with no Auto-Trigger
    // handle, enable, source, treshold, direction, delay (value count), autotrigger (ms)
    pico_error = ps5000aSetSimpleTrigger (pico_handle, 1, PS5000A_EXTERNAL, 0, PS5000A_RISING, 0, 0);
    CHECK_PICO_ERROR ("couldn't set Trigger");

    // allocate memory for the input data buffers --> BUFFER_SIZE*2 : two byte per item
    pico_buffer_A = malloc (BUFFER_SIZE*sizeof(pico_buffer_A[0]));
    pico_buffer_B = malloc (BUFFER_SIZE*sizeof(pico_buffer_B[0]));
    pico_buffer_C = malloc (BUFFER_SIZE*sizeof(pico_buffer_C[0]));
    pico_buffer_D = malloc (BUFFER_SIZE*sizeof(pico_buffer_D[0]));
    if (pico_buffer_A == 0 || pico_buffer_B == 0 || pico_buffer_C == 0 || pico_buffer_D == 0) {
        Tcl_SetResult (interp, "picoScope: couldn't allocate buffer memory.", TCL_VOLATILE);
        return TCL_ERROR;
    }

    // tell the picoscope library from the allocated data buffers
    pico_error = ps5000aSetDataBuffer (pico_handle, PS5000A_CHANNEL_A, pico_buffer_A,
                                       BUFFER_SIZE, 0,  PS5000A_RATIO_MODE_AVERAGE);
    CHECK_PICO_ERROR ("couldn't set Data Buffer for channel A");

    pico_error = ps5000aSetDataBuffer (pico_handle, PS5000A_CHANNEL_B, pico_buffer_B,
                                       BUFFER_SIZE, 0,  PS5000A_RATIO_MODE_AVERAGE);
    CHECK_PICO_ERROR ("couldn't set Data Buffer for channel B");

    pico_error = ps5000aSetDataBuffer (pico_handle, PS5000A_CHANNEL_C, pico_buffer_C,
                                       BUFFER_SIZE, 0,  PS5000A_RATIO_MODE_AVERAGE);
    CHECK_PICO_ERROR ("couldn't set Data Buffer for channel C");

    pico_error = ps5000aSetDataBuffer (pico_handle, PS5000A_CHANNEL_D, pico_buffer_D,
                                       BUFFER_SIZE, 0,  PS5000A_RATIO_MODE_AVERAGE);
    CHECK_PICO_ERROR ("couldn't set Data Buffer for channel D");

    // allocate memory for the AWG data buffer
    awg_buffer = malloc (AWG_SIZE*sizeof(awg_buffer[0]));
    if (awg_buffer == 0) {
        Tcl_SetResult (interp, "picoScope: couldn't allocate AWG buffer memory.", TCL_VOLATILE);
        return TCL_ERROR;
    }

    awg_buffer_double = malloc (AWG_SIZE*sizeof(awg_buffer_double[0]));
    if (awg_buffer_double == 0) {
        Tcl_SetResult (interp, "picoScope: couldn't allocate internal AWG buffer memory.", TCL_VOLATILE);
        return TCL_ERROR;
    }

    return TCL_OK;
}

int picoScopeEndCmd (ClientData clientData, Tcl_Interp *interp,
                     int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1) {
        Tcl_WrongNumArgs(interp, 1, objv, "");
        return TCL_ERROR;
    }

    if (pico_handle == 0) {
        Tcl_SetResult (interp, "picoScope: wasn't initialized", TCL_VOLATILE);
        return TCL_ERROR;
    }

    pico_error = ps5000aCloseUnit (pico_handle);
    pico_handle = 0;
    CHECK_PICO_ERROR ("problem during closing the USB connection");

    // free all allocated memory
    free (pico_buffer_A);
    free (pico_buffer_B);
    free (pico_buffer_C);
    free (pico_buffer_D);
    free (awg_buffer);

    pico_buffer_A = 0;
    pico_buffer_B = 0;
    pico_buffer_C = 0;
    awg_buffer = 0;

    return TCL_OK;
}

int picoScopeTriggerCmd (ClientData clientData, Tcl_Interp *interp,
                         int objc, Tcl_Obj *CONST objv[])
{
    char *trigger_type;

    if (pico_handle == 0) {
        Tcl_SetResult (interp, "picoScope: isn't initialized", TCL_VOLATILE);
        return TCL_ERROR;
    }

    if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "?triggertype?");
        return TCL_ERROR;
    }

    trigger_type = Tcl_GetString (objv[1]);

    if (strcmp ("AUTO", trigger_type) == 0) {
        // Set Trigger to Auto - 100 ms delay
        // Trigger - external, rising edge, with Auto-Trigger (100 ms)
        // handle, enable, source, treshold, direction, delay (value count), autotrigger (ms)
        pico_error = ps5000aSetSimpleTrigger (pico_handle, 1, PS5000A_EXTERNAL, 0, PS5000A_RISING, 0, 100);
        CHECK_PICO_ERROR ("couldn't set Trigger");
    } else if (strcmp ("NORM", trigger_type) == 0) {
        // Set Trigger to Norm (no Auto...)
        pico_error = ps5000aSetSimpleTrigger (pico_handle, 1, PS5000A_EXTERNAL, 0, PS5000A_RISING, 0, 0);
        CHECK_PICO_ERROR ("couldn't set Trigger");
    } else {
        snprintf (err_msg, sizeof(err_msg), "picoScopeTrigger: wrong trigger type - "
                  "expected AUTO or NORM, got '%s'", trigger_type);
        Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
        return TCL_ERROR;
    }

    return TCL_OK;
}


int picoScopeAwgCmd (ClientData clientData, Tcl_Interp *interp,
                     int objc, Tcl_Obj *CONST objv[])
{
    int frequency;
    double amplitude, awg_output_amplitude;
    int i, j, entries_written;
    FILE *awg_file;
    double delta;

    if (pico_handle == 0) {
        Tcl_SetResult (interp, "picoScope: isn't initialized", TCL_VOLATILE);
        return TCL_ERROR;
    }

    if ((objc != 2) && (objc != 3) && (objc != 5) && (objc != 7) && (objc != 9) && (objc != 11)) {
        Tcl_WrongNumArgs(interp, 1, objv, "?single argument 'OFF' or one to "
                         "five frequency [kHz, integer] / amplitude [V] pairs?");
        return TCL_ERROR;
    }

    if (objc == 2) {
        if (strcmp ("OFF", Tcl_GetString (objv[1])) != 0) {
            snprintf (err_msg, sizeof(err_msg), "picoScopeAwg: wrong 'single' argument value - "
                      "expected OFF, got '%s'", Tcl_GetString(objv[1]));
            Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
            return TCL_ERROR;
        } else {
            // Turn off AWG by writting negative Maximum into the buffer -->
            // this drives the power amplifier into negative saturation
            // minimizing its current consumption and therefore temperature...

            awg_output_amplitude = 2.0;
            for (i=0; i<AWG_SIZE; i++) {
                awg_buffer[i] = -32768;
            }
        }
    } else {
        // I got some frequency/amplitude pairs...

        // awg_output * 6 shall be smaller than 10 V --> else input amplifiers could be overdriven...
        awg_output_amplitude = 1.6;

        for (i=0; i<AWG_SIZE; i++) {
            awg_buffer_double[i] = 0.0;
        }

        for (j=1; j<objc; j+=2) {
            if (Tcl_GetIntFromObj (interp, objv[j], &frequency) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Tcl_GetDoubleFromObj (interp, objv[j+1], &amplitude) != TCL_OK) {
                return TCL_ERROR;
            }

            if ((frequency < 30) || (frequency > 2000)) {
                snprintf (err_msg, sizeof(err_msg),
                          "picoScopeAwg: frequency (arg %i) out of range - "
                          "expected 30-2000 kHz, got %i kHz", j, frequency);
                Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
                return TCL_ERROR;
            }

            if ((amplitude < 0.1) || (amplitude > 2.0)) {
                snprintf (err_msg, sizeof(err_msg),
                          "picoScopeAwg: amplitude (arg %i) out of range - "
                          "expected 0.1 - 2 V, got %lf V", j+1, amplitude);
                Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
                return TCL_ERROR;
            }

            for (i=0; i<AWG_SIZE; i++) {
                awg_buffer_double[i] += amplitude * sin (2*M_PI*frequency*i/AWG_SIZE);
            }
        }

        // check for buffer overflow and copy data to output buffer

        for (i=0; i<AWG_SIZE; i++) {
            if (abs(awg_buffer_double[i]) > awg_output_amplitude) {
                snprintf (err_msg, sizeof(err_msg),
                          "picoScopeAwg: amplitude overflow - "
                          "check that the sum of amplitudes is lower than %.3f V",
                          awg_output_amplitude);
                Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
                return TCL_ERROR;
            }

            // add noise to eliminate harmonic distortions... !?
            //awg_buffer[i] = round (awg_buffer_double[i] / 2.0 * 32757 + random()/RAND_MAX*10.0);

            awg_buffer[i] = round (awg_buffer_double[i] / 2.0 * 32767);
        }
    }

    // Write the actual awg-data into a file - but only when doing frequency/value pairs...
    if (objc > 2) {
        awg_file = fopen ("awg_data.dat", "w");
        if (awg_file == 0) {
            Tcl_SetResult (interp, "picoScopeAwg: couldn't open datafile "
                           "'awg_data.dat' for writing", TCL_VOLATILE);
            return TCL_ERROR;
        }

        entries_written = fwrite (awg_buffer, sizeof(awg_buffer[0]), AWG_SIZE, awg_file);
        if (entries_written != AWG_SIZE) {
            snprintf (err_msg, sizeof(err_msg), "picoScopeAwg: couldn't write "
                      "'awg_data.dat' file - wrote only %i from %i entries.\n",
                      entries_written, AWG_SIZE);
            Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
            return TCL_ERROR;
        }

        if (fclose (awg_file) == EOF) {
            Tcl_SetResult (interp, "picoScopeAwg: file close of 'awg_data.dat' failed", TCL_VOLATILE);
            return TCL_ERROR;
        }
    }

    // Send the awg-data to the picoscope

    /* aus Forum
       delta = ((1.0 * frequency * waveformSize) / PS5X44A_MAX_SIG_GEN_BUFFER_SIZE)
       * (AWG_PHASE_ACCUMULATOR * 1/AWG_DAC_FREQUENCY);
       bzw, mit Zahlen:
       delta = frequency * waveformSize / 49152 (bzw. 48k) * 2^32 / 200MHz
    */

    // Buffer soll mit einer Wiederholrate von 1 kHz ausgegeben werden...
    delta = round(1e3 * AWG_SIZE / 49152.0 * pow (2.0, 32.0) / 200e6);

    // maximum amplitude
    amplitude = 2;

    pico_error = ps5000aSetSigGenArbitrary (
        pico_handle, 0, round(amplitude*2*1e6), delta, delta, 0, 10,
        awg_buffer, AWG_SIZE, PS5000A_UP, PS5000A_ES_OFF,
        PS5000A_SINGLE, 0, 0, PS5000A_SIGGEN_RISING, PS5000A_SIGGEN_NONE, 0);
    CHECK_PICO_ERROR ("couldn't program AWG");

    return TCL_OK;
}

int picoScopeMeasureCmd (ClientData clientData, Tcl_Interp *interp,
                         int objc, Tcl_Obj *CONST objv[])
{
    double duration;
    unsigned number_of_samples, number_of_samples_out;
    int number_of_pre_trigger_samples;
    int i, wait_time;
    int16_t pico_ready;
    int16_t overflow;
//    struct timeval zeit_start, zeit_end;
//    double transfer_time;

    FILE *datafile;
    int items_written;

    if (pico_handle == 0) {
        Tcl_SetResult (interp, "picoScope: isn't initialized", TCL_VOLATILE);
        return TCL_ERROR;
    }

    if (objc != 3) {
        Tcl_WrongNumArgs(interp, 1, objv, "?measurement duration? ?filename?");
        return TCL_ERROR;
    }

    if (Tcl_GetDoubleFromObj (interp, objv[1], & duration) != TCL_OK) {
        return TCL_ERROR;
    }

    if ((duration < 0.0) || (duration > 0.536)) {
        snprintf (err_msg, sizeof(err_msg), "picoScope: measurement duration "
                  "out of bounds (0.0 - 0.536 s): got %lf", duration);
        Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
        return TCL_ERROR;
    }

    // Do the measurement - Time Base 3: 125 MS/s

    // preTriggerSamples: 0.5 ms - so the first 1 ms frame is "around" the
    // zero crossing of the mains voltage
    number_of_pre_trigger_samples = round(125e6*0.5e-3);

    number_of_samples = ceil (125e6 * duration);
    pico_error = ps5000aRunBlock (pico_handle, number_of_pre_trigger_samples,
                                  number_of_samples - number_of_pre_trigger_samples,
                                  3, 0, 0, 0, 0);
    CHECK_PICO_ERROR ("couldn't do `RunBlock`");


    // Maximum wait time in multiples of 10 ms --> max auto trigger time + acquisition time
    // +10 to be sure I wait long enough without throwing an unnecessary error
    wait_time = (int) (30 + ceil ((50e-3 + duration)/10e-3));

    for (i=0; i<wait_time; i++) {
        pico_error = ps5000aIsReady (pico_handle, &pico_ready);
        CHECK_PICO_ERROR ("couldn't get ready-status");

        if (pico_ready != 0) {
            // data ready
            break;
        }

        // Sleep 10 ms and try again...
        usleep (10000);
    }

    if (i==wait_time) {
        Tcl_SetResult (interp, "picoScope: timeout during measurement", TCL_VOLATILE);
        return TCL_ERROR;
    }

    // Get values from picoscope --> downsample ratio 1:25
//    gettimeofday (&zeit_start, 0);
    number_of_samples = floor (number_of_samples / 25);
    number_of_samples_out = number_of_samples;
    pico_error = ps5000aGetValues (pico_handle, 0, &number_of_samples_out, 25,
                                   PS5000A_RATIO_MODE_AVERAGE, 0, &overflow);
    CHECK_PICO_ERROR ("couldn't get values from picoscope");

    if (number_of_samples != number_of_samples_out) {
        snprintf (err_msg, sizeof(err_msg), "picoScope: didn't get the %i requested samples - "
                  "got %i", number_of_samples, number_of_samples_out);
        Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
        return TCL_ERROR;
    }

//    gettimeofday (&zeit_end, 0);
//    transfer_time = (zeit_end.tv_sec - zeit_start.tv_sec) + (zeit_end.tv_usec-zeit_start.tv_usec)*1e-6;

    datafile = fopen (Tcl_GetString (objv[2]), "w");
    if (datafile == 0) {
        Tcl_SetResult (interp, "picoScopeMeasure: couldn't open datafile "
                       "'picoscope.dat' for writing", TCL_VOLATILE);
        return TCL_ERROR;
    }

    items_written = 0;
    items_written += fwrite (pico_buffer_A, sizeof(pico_buffer_A[0]), number_of_samples, datafile);
    items_written += fwrite (pico_buffer_B, sizeof(pico_buffer_B[0]), number_of_samples, datafile);
    items_written += fwrite (pico_buffer_C, sizeof(pico_buffer_C[0]), number_of_samples, datafile);
    items_written += fwrite (pico_buffer_D, sizeof(pico_buffer_D[0]), number_of_samples, datafile);

    if (items_written != 4*number_of_samples) {
        snprintf (err_msg, sizeof(err_msg), "picoScopeMeasure: couldn't write "
                  "'picoscope.dat' file - wrote only %i from %i entries.\n",
                  items_written, 4*number_of_samples);
        Tcl_SetResult (interp, err_msg, TCL_VOLATILE);
        return TCL_ERROR;
    }

    if (fclose (datafile) == EOF) {
        Tcl_SetResult (interp, "picoScope: file close of 'picoscope.dat' failed", TCL_VOLATILE);
        return TCL_ERROR;
    }

    return TCL_OK;
}
